/** 
  *  \file CplxFFT3.hxx
  *  Capsule for performing complex Fast Fourier Transforms
  *    
  *  \author G. Martin <graemem@ssec.wisc.edu>
  *
  *  \version $Id: CplxFFT3.hxx,v 1.2.2.7 2006/03/29 16:53:44 maciek Exp $
  *
  *  \par Copyright:
  *  \verbatim
  *
  *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2006
  *  Space Science and Engineering Center
  *  University of Wisconsin - Madison, USA
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  \endverbatim
**/


#ifndef H_CPLXFFT3
#define H_CPLXFFT3

// STL includes
#include <complex>
#include "CplxUtils.hxx"

// external lib includes
#include "fftw3.h"


namespace gips
{      
    /*!  functor for use with std::transform checks if values are finite.
      *
      *  \par Requirements:
      *  give it a bool pointer, check its value after the copy.
      *
      *  \par Restrictions:
      *  must have OutType be able to create an object with an instance of InType.
      */
    template <typename InType, typename OutType>
    class transformWithFiniteCheck 
    {
    private:
        bool *flag; 

    public:

        // ctor: pass in pointer to external flag that will be set to true
        // when a non-finite value is found
        transformWithFiniteCheck(bool* flag_): flag(flag_) {}

        static bool isFinite(std::complex<double> d) 
            { 
                return std::isfinite(d.real()) && std::isfinite(d.imag()); 
            }
        static bool isFinite(double d) 
            { 
                return std::isfinite(d); 
            }
        
        // complex ints are always finite
        static bool isFinite(std::complex<int> i) 
            { 
                return true; 
            }
        
        OutType operator()(const InType& x )
            {
                if(!isFinite(x)) *flag = true;
                return OutType(x);
            }
    };  // transformWithFiniteCheck

    /*! Complex Fast Fourier Transform class, acts as a resource wrapper around the blazingly
      * fast FFTW3 library.
      *
      *  \par Theory: 
      *  Easy to use: initialize with a couple of options, then provide input and output buffers.
      *  Internally, the buffers are copied to private, specially allocated FFTW buffers.
      *
      *  \par Requirements:
      *  FFT direction is determined by the client according to algorithmic needs.
      *  training (measure vs estimate) is a performance tradeoff: fast startup with estimate,
      *  vs better performance-per-fft with measure.

      *  \par Restrictions:
      *  InType and OutType need to be statically sized arrays of the same size
      *
      */
    template <typename InType, typename OutType>
    class CplxFFT
    {
    public:
        enum direction_e { FORWARD = FFTW_FORWARD, BACKWARD = FFTW_BACKWARD };
        enum training_e { MEASURE = FFTW_MEASURE, ESTIMATE = FFTW_ESTIMATE };
        
        struct settings_t
        {
            direction_e fftDirection;
            training_e fftTraining;
            //explicit settings_t(unsigned int is_, direction_e fd_=FORWARD, training_e ft_=MEASURE): 
            //    fftDirection(fd_), fftTraining(ft_)
            //{}

            template< typename Src >
            static settings_t fromStruct( const Src &x )
            {
                settings_t ret;
                ret.fftDirection = direction_e(x.fftDirection);
                ret.fftTraining = training_e(x.fftTraining);
                return ret;
            }
                    
            
        }; 

        struct Ports
        {
            const InType* obs_in;
            OutType* obs_out;

            Ports(const InType *obs_in_, OutType *obs_out_): 
                    obs_in(obs_in_), obs_out(obs_out_) {}
        };
        
    private:
        const settings_t settings;
        Ports ports;

        fftw_complex *in, *out;
        fftw_plan plan;
        
    public:
        CplxFFT(const settings_t& s_, const Ports& p_):
            settings(s_), ports(p_), 
                in((fftw_complex*)fftw_malloc(sizeof(fftw_complex) * ports.obs_in->size())),
                out((fftw_complex*)fftw_malloc(sizeof(fftw_complex) * ports.obs_in->size())),
                plan(
                    fftw_plan_dft_1d
                    (
                        ports.obs_in->size(), 
                        in, out, 
                        settings.fftDirection, settings.fftTraining
                    )
                ) 
        {
            assert(ports.obs_in->size() == ports.obs_out->size());
        }

        ~CplxFFT()
        {
            fftw_destroy_plan(plan);
            fftw_free(in);
            fftw_free(out);
        } 
        
        void operator()()
        {
            bool flag = false;
            transformWithFiniteCheck<typename InType::value_type, std::complex<double> > twfc(&flag);
                    
            std::transform( ports.obs_in->begin(), 
                            ports.obs_in->end(), 
                            (std::complex<double>*)in, 
                            twfc);

            // if a non-finite value was encountered, bypass FFT and set output
            // buffer to all oob values
            if(flag) 
            {
                typename OutType::value_type oob(oobVal<typename OutType::value_type>());
                std::fill( ports.obs_out->begin(), ports.obs_out->end(), oob);
                return;
            }

            fftw_execute(plan);
 
            std::copy(
                (std::complex<double>*)out, 
                (std::complex<double>*)out + ports.obs_out->size(), 
                ports.obs_out->begin()
            );
        }
    }; // CplxFFT
}

#endif
