/*
 * \brief SHIS implementation of Ifg2SpectrumStage
 * \author Maciek Smuga-Otto <maciek@ssec.wisc.edu>
 *
 * based directly on the Scanning-HIS rsh2fbf application, written by Ray Garcia <rayg@ssec.wisc.edu>
 *
 *  \version $Id: IfgToSpectrum.hxx,v 1.6.2.3 2006/03/30 21:04:45 graemem Exp $
 *
 *  \par Copyright:
 *  \verbatim
 *
 *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
 *  Space Science and Engineering Center
 *  University of Wisconsin - Madison, USA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  \endverbatim
 */

// NOTE: THIS FILE WILL BE AUTO-GENERATED IN THE FUTURE FROM INTEFACE DESCRIPTIONS
// OF THE SUBCAPSULES BY MEANS OF A FLANGING COMPILER

#include <complex>
//#include <vector>

#include <boost/shared_ptr.hpp>
#include <boost/array.hpp>

#include "IfgShift.hxx"
#include "CplxFFT3.hxx"
#include "SpectrumChop.hxx"

#include "CplxUtils.hxx"  // debugging utils
#include "FBF.h"

namespace gips {

template
<
    typename InType,
    typename OutType 
>
class IfgToSpectrum
{
    // internal sub-capsule types
    typedef boost::array< std::complex< double >, InType::static_size > buffer; // internal buffer type
    
    typedef IfgShift<InType,buffer> MyIfgShift;
    typedef CplxFFT<buffer,buffer> MyCplxFFT;
    typedef SpectrumChop<buffer,OutType> MySpectrumChop;

#ifdef DEBUG_MODE
    static const bool debugout = true;
#else
    static const bool debugout = false;
#endif    

public:
    enum direction_e { FORWARD = MyCplxFFT::FORWARD, BACKWARD = MyCplxFFT::BACKWARD }; // FUTURE make this private, remove with fromStructs
    enum training_e { MEASURE = MyCplxFFT::MEASURE, ESTIMATE = MyCplxFFT::ESTIMATE };
             
    struct settings_t 
    {
        unsigned int chopRawSpectrumFrom;  
        direction_e fftDirection; 
        training_e fftTraining;
        bool wrapRawSpectrumLastPoint; 

        static bool doWrapPoint()
        {
            // wrap last point iff spectrum size is odd.
            return OutType::static_size % 2 == 1;
        }

        settings_t(unsigned chopRawSpectrumFrom_, 
                   training_e fftTraining_) : 
            chopRawSpectrumFrom(chopRawSpectrumFrom_),
            fftDirection(FORWARD),
            fftTraining(fftTraining_),
            wrapRawSpectrumLastPoint(doWrapPoint()) {}

        settings_t() {}

        template< typename Src > 
        static settings_t fromStruct( const Src &x )
        {
            settings_t ret;
            ret.chopRawSpectrumFrom = x.chopRawSpectrumFrom;
            ret.fftTraining = x.fftTraining;
            // FUTURE - the below should be a function of the spectrum being even or odd sized
            ret.wrapRawSpectrumLastPoint = x.wrapRawSpectrumLastPoint;
            return ret;
        }
    };
    
    struct Ports 
    {
        const InType *obs_in;
        OutType *obs_out;
                
        Ports(const InType *obs_in_,
              OutType *obs_out_):   
           obs_in(obs_in_), obs_out(obs_out_) { }
    };

protected:
    settings_t settings;
    Ports P;
    
    buffer is2cf;
    buffer cf2sc;

    MyIfgShift capsule1;
    MyCplxFFT capsule2;
    MySpectrumChop capsule3;
    typedef typename MyCplxFFT::settings_t CplxSettings;    
    CplxSettings _cfSettings()
    {
        CplxSettings s(CplxSettings::fromStruct( settings ));
        s.fftDirection = MyCplxFFT::FORWARD;
        return s;
    }
    
    typename MySpectrumChop::settings_t _scSettings()
    {
        return MySpectrumChop::settings_t::fromStruct(settings);
    }
        
public:
    explicit IfgToSpectrum( const settings_t &settings_, const Ports &ports ): settings(settings_), P(ports),
        // initialize sub-capsules to move data between capsule input and output
        capsule1( MyIfgShift::Ports(ports.obs_in,&is2cf) ), 
        capsule2( _cfSettings(), MyCplxFFT::Ports(&is2cf,&cf2sc) ), 
        capsule3( _scSettings(), MySpectrumChop::Ports(&cf2sc,ports.obs_out) )
    { }  
    
    void operator() ()
    {   
        if (debugout)
            FBF::save("ifg.real8.2048", *(P.obs_in));
        capsule1();
        if (debugout)
        {
            boost::array<double, buffer::static_size> debugvec;
            extractComplexAbs(is2cf,debugvec);
            FBF::save("is2cfAbs.real8.2048",  debugvec);
            extractComplexPhase(is2cf,debugvec);
            FBF::save("is2cfPhase.real8.2048",  debugvec);
        }
             
        capsule2();
        if (debugout)
        {
            boost::array<double, buffer::static_size> debugvec;
            extractComplexAbs(cf2sc,debugvec);
            FBF::save("cf2scAbs.real8.2048",  debugvec);
            extractComplexPhase(cf2sc,debugvec);
            FBF::save("cf2scPhase.real8.2048",  debugvec);
        }
         
        capsule3();
        if (debugout)
        {
            boost::array<double, OutType::static_size> debugvec;
            extractComplexAbs(*(P.obs_out),debugvec);
            FBF::save("cplxspecAbs.real8.1024",  debugvec);
            extractComplexPhase(*(P.obs_out),debugvec);
            FBF::save("cplxspecPhase.real8.1024",  debugvec);
        }
         
    }
    
}; // class
    
} // Namespace
