/** 
*  \file SimCal3.hxx
*  \brief Pipeline capsule for GIFTS data calibration
*  \verbatim    
* OVERVIEW
*
* This capsule contains the main data processing pipeline. Input interferograms
* are pushed through either the STUDY sub-pipeline (H, A, Z blackbody scenes) 
* or the APPLY sub-pipeline (E scenes). The study sub-pipeline loads the 
* calibration context cache with context instances which can be used to 
* calibrate earth data:
*    
*    IfgToSpectrum -> PiecewiseLinearCalibrationStudy
*
* The apply sub-pipeline uses the data available in the context cache in order 
* to synthesize (interpolate in this case) the blackbody radiance data which 
* feeds the 3-body calibration equation:
*
*    (IfgToSpectrum; ThreeBodyPolynomialCalibrationSynthesis) -> 
*      ThreeBodyRadiometricCalibration -> CosineRolloff -> Resampler
*
* INTERFACE
*
*   Ports:
*   -- ancillary: ancillary data struct (per-invocation data)
*   -- ref: reference data struct (can change during capsule lifetime))
*   -- control: control data struct (determines invocation operation)
*   -- obs: input data buffer - raw interferogram 
*   -- emiss: input data buffer - emissivity
*   -- obsSpectrum: output data buffer - uncalibrated spectrum after 
*      IfgToSpectrum
*   -- calSpectrum: output data buffer - calibrated spectrum, after 
*      CosineRolloff
*   -- outSpectrum: output data buffer - resampled 
*
*   Ancillary:
*   -- obsType: observation type (H, A, Z, E)
*   -- scanDir: scan direction (FORWARD, BACKWARD)
*   -- time: observation time in Unix epoch time units
*   -- hbbTemp: hot blackbody temperature in degrees K
*   -- abbTemp: warm blackbody temperature in degrees K
*   -- zbbTemp: space temperature in degrees K
*   -- reflectedTemp: temperature of the optical system in degrees K
*   -- detectorId: index of detector identifying position in detector array
*
*   Reference:
*   -- theta: angle in milliradians of the center of the detector pixel from the optical axis
*   -- telescopeTau: reflectivity of telescope
*   -- calMirrorTau: reflectivity of calibration mirror
*
*   Control:
*   -- operation: specifies invocation operation (process data or reset cache)
*
*   Settings (passed to ctor):
*   -- wnumScale: target wavenumber scale
*   -- rolloffWn0, rolloffWn1, rolloffWn2, rolloffWn3: wavenumbers to use in 
*      rolloff
*
* TEMPLATE SPECIFICATION 
*   
*   Parameters:
*   -- InType: type of input buffer obs
*   -- EmissInType: type of input buffer emiss
*   -- ObsOutType: type of output buffer obsSpectrum
*   -- OutType: type of output buffers calSpectrum and outSpectrum 
*   -- RESAMPLE_UPDSAMPLED_SIZE: size of upsampled buffer to use in Resampler
*
*   Requirements:      
*   -- Buffer types must be statically sized array-like (for example boost::array),  
*      with the following capabilities: begin(), end(), size(), operator[], 
*      value_type, static_size
*   -- Input buffer value_types must be (auto) convertible to output buffer value_types  
*      and compatible with type int, and must support arithmetic operators 
*      ('/', '*', '+', '-') 
*   -- InType::static_size == (2 * OutType::static_size)
*      -- FUTURE: this restriction may be removed in the future to deal with real GIFTS data.
*  \endverbatim
*
*  \author G. Martin <graemem@ssec.wisc.edu>
*
*  \version $Id: SimCal3.hxx,v 1.1.4.11 2006/04/04 19:00:44 graemem Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2006
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
*  \endverbatim
*/

#ifndef HXX_SIMCAL3
#define HXX_SIMCAL3

#include <algorithm>
#include <cmath>

#include <boost/array.hpp>
#include <boost/static_assert.hpp>

#include "Physics.h"                              // WavenumberScale<>, DEBUG_MODE
#include "IfgToSpectrum.hxx"                      // IfgToSpectrum<>
#include "PiecewiseLinearCalibrationContext.hxx"  // PiecewiseLinearCalibrationStudy<>, 
                                                  // PolynomialCalibrationContext, 
                                                  // ThreeBodyPolynomialCalibrationSynthesis<>
#include "RadiometricCalibration.hxx"             // ThreeBodyRadiometricCalibration<>
#include "CosineRolloff.hxx"                      // CosineRolloff<>
#include "Resample.hxx"                           // SpectralResampling<>

#include "FBF.h"   // DEBUG output 
#include <complex>

namespace gips 
{

    template< typename InType, typename EmissInType, typename ObsOutType, 
              typename OutType, unsigned RESAMPLE_UPSAMPLED_SIZE = 262144 >
    class SimCal3
    {
        static const std::size_t IFG_SIZE = InType::static_size;
        static const std::size_t SPC_SIZE = OutType::static_size;

        BOOST_STATIC_ASSERT((SPC_SIZE*2 == IFG_SIZE));
        
        typedef boost::array< std::complex<double>, SPC_SIZE > CplxSpectrum;
        typedef boost::array< double, SPC_SIZE > RealSpectrum;
         
        typedef IfgToSpectrum< InType, CplxSpectrum > Ifg2Spc;
        typedef PolynomialCalibrationContext::BasicCacheType ContextCache;
        typedef PiecewiseLinearCalibrationStudy< CplxSpectrum, ContextCache > CalStudy;
        typedef ThreeBodyPolynomialCalibrationSynthesis< ContextCache, CplxSpectrum > CalSynth;
        // three-body calibration includes as reference an emissivity spectrum
        typedef ThreeBodyRadiometricCalibration< CplxSpectrum, RealSpectrum, physics::WnumScale<double,SPC_SIZE>, RealSpectrum > CalThreeBody;
        // FUTURE - put downsampler module before the above line
        typedef CosineRolloff< RealSpectrum, RealSpectrum > Rolloff;
        typedef SpectralResampling<RealSpectrum, RealSpectrum, RESAMPLE_UPSAMPLED_SIZE> Resample;

    public:   

        // determines operation performed on capsule invocation
        enum operation_e
        {
            RESET_BB_CACHE = 0,      // reset the blackbody cache
            PROCESS_DATA = 1         // normal operation
        };
    
        struct ancillary_t
        {
            physics::observation_type_e obsType;
            physics::scan_direction_e scanDir;
            double time;
            double hbbTemp;
            double abbTemp;
            double zbbTemp;
			double reflectedTemp;
            unsigned detectorId;
        };
 
        struct settings_t
        {
//            unsigned optimization_level;
//            unsigned verbosity;

            physics::WnumScale<double, SPC_SIZE> wnumScale;

            double rolloffWn0;
            double rolloffWn1;
            double rolloffWn2;
            double rolloffWn3;

            settings_t(const physics::WnumScale<double, SPC_SIZE> wnumScale_, 
                       const double rolloffWn0_, const double rolloffWn1_, 
                       const double rolloffWn2_, const double rolloffWn3_):
                wnumScale(wnumScale_), rolloffWn0(rolloffWn0_), 
                rolloffWn1(rolloffWn1_), rolloffWn2(rolloffWn2_),
                rolloffWn3(rolloffWn3_){}
        };
		
		struct ref_t
		{
			double theta;
			double telescopeTau;
			double calMirrorTau;
		};

        struct control_t
        {
            operation_e operation;
        };

        struct Ports
        {
            const ancillary_t *ancillaryPtr;
            const ref_t *refPtr;
            const control_t *controlPtr;
            const InType *obsPtr;
            const EmissInType *emissPtr;
            ObsOutType *obsSpectrumPtr;
            OutType *calSpectrumPtr;
            OutType *outSpectrumPtr;
            
            Ports(const ancillary_t* anc_, const ref_t* ref_, const control_t* control_, 
                  const InType* in_, const EmissInType* emiss_, 
                  ObsOutType *obsOut_, OutType* cal_, OutType* out_):
                ancillaryPtr(anc_), refPtr(ref_), controlPtr(control_), 
                obsPtr(in_), emissPtr(emiss_), obsSpectrumPtr(obsOut_), 
                calSpectrumPtr(cal_), outSpectrumPtr(out_) {}
        };
        
    private:
        physics::WnumScale<double, SPC_SIZE> targetWnumScale;
        const Ports ports;

        // internal buffers
        const InType &obs;                           // set to *ports.obsPtr    
        ObsOutType &obsSpectrum;                     // set to *ports.obsSpectrumPtr         
        CplxSpectrum abbSynth;
        CplxSpectrum zbbSynth;
        CplxSpectrum hbbSynth;
        RealSpectrum cal;          
        RealSpectrum diag;
        RealSpectrum &rolledCal;           // set to *ports.calSpectrumPtr // FUTURE: consolidate this with cal buffer, rolloff in place
        RealSpectrum &resampledCal;            // set to *ports.outSpectrumPtr

        // subcapsule ref structs
        typename CalThreeBody::reference_t calRef;
        
        // subcapsule ancillary structs
        typename CalStudy::ancillary_t studyAncil;        
        typename CalSynth::ancillary_t synthAncil;
        typename CalThreeBody::ancillary_t calAncil;
        typename Rolloff::ancillary_t rollAncil;
        typename Resample::ancillary_t rpleAncil;

        // subcapsule ports
        typename Ifg2Spc::Ports i2sPorts;                  
        typename CalStudy::Ports studyPorts;
        typename CalSynth::Ports synthPorts;
        typename CalThreeBody::Ports calPorts;
        typename Rolloff::Ports rollPorts;
        typename Resample::Ports rplePorts;
   
        // context cache holds the data needed to synthesize time-matched blackbody "observations"
        ContextCache contexts;
  
        // subcapsules
        Ifg2Spc i2s;          // Ifg to spectrum
        CalStudy study;       // Cal study
        CalSynth synth;       // Cal synth
        CalThreeBody c3b;     // Three body cal
        Rolloff rolloff;      // Rolloff 
        Resample resample;    // Spectral Resampling

        // FUTURE: promote or replace 
        int round(double d) { return (d>0 ? int(d+0.5) : int(d-0.5)); }

     public:
        
        // ctor 
        // initializer list: initialize own members, internal buffers, subcapsule 
        // settings, subcapsule ports, subcapsules. 
        SimCal3(const settings_t settings_, const Ports &p_): 
            targetWnumScale(settings_.wnumScale),
            ports(p_),  
            
			// boundary buffers
            obs(*ports.obsPtr),                   // obs in
            obsSpectrum(*ports.obsSpectrumPtr),  // obs spectrum out
            rolledCal(*ports.calSpectrumPtr),             // cal spectrum out
            resampledCal(*ports.outSpectrumPtr),    // resampled spectrum out

            // subcapsules' ports
            i2sPorts(&obs, &obsSpectrum ),          
            studyPorts( &obsSpectrum, &studyAncil, &contexts ),
            synthPorts( &synthAncil, &contexts, 
                        &hbbSynth, &abbSynth, &zbbSynth ),
            calPorts( &calRef, &calAncil, &obsSpectrum, 
                      &hbbSynth, &abbSynth, &zbbSynth, &cal, &diag ),
            rollPorts(&rollAncil, &cal, &rolledCal),
            rplePorts(&rpleAncil, &rolledCal, &resampledCal),

            // subcapsules
            // FUTURE: set training_e from debug level passed in through settings
            // MEASURE = delayed startup, fast per pixel,
            // ESTIMATE = slower per pixel performance. Set to MEASURE for prod
			// and ESTIMATE for debug( also fbf file output, fewer pixels)
#ifdef DEBUG_MODE
            i2s(Ifg2Spc::settings_t(unsigned(IFG_SIZE/2), Ifg2Spc::ESTIMATE), i2sPorts),
#else
            i2s(Ifg2Spc::settings_t(unsigned(IFG_SIZE/2), Ifg2Spc::MEASURE), i2sPorts),
#endif
            study(studyPorts),
            synth(synthPorts),
            c3b(calPorts),
            rolloff(rollPorts),
            resample(Resample::settings_t(settings_.wnumScale), rplePorts)   
            {
                // Rolloff subcapsule's ancil structs: set here since they 
                // will not change per pixel
                rollAncil.p0 = round(settings_.wnumScale.indexOf(settings_.rolloffWn0));
                rollAncil.p1 = round(settings_.wnumScale.indexOf(settings_.rolloffWn1)); 
                rollAncil.p2 = round(settings_.wnumScale.indexOf(settings_.rolloffWn2)); 
                rollAncil.p3 = round(settings_.wnumScale.indexOf(settings_.rolloffWn3));
            }   // ctor
        
        ~SimCal3() {}
        void operator()() 
            {  
                // convenience references
                const ancillary_t &anc(*ports.ancillaryPtr);
                const ref_t &ref(*ports.refPtr);                    
                const control_t &control(*ports.controlPtr);

                switch(control.operation)
                {
                case RESET_BB_CACHE:
                    contexts.clear();
                    return;
                case PROCESS_DATA:
                    break;
                default:
                    abort();
                }

                i2s();

                switch(anc.obsType)
                {
                // if blackbody: run through study pipeline
                case physics::OBS_REF_UPPER:
                case physics::OBS_REF_LOWER:
                case physics::OBS_REF_SPACE:
                    studyAncil.time = anc.time;
                    studyAncil.detectorid = anc.detectorId;
                    studyAncil.scenetype = anc.obsType;
                    studyAncil.scandir = anc.scanDir;
 
                    study();
                    break;
                // if earth scene, run through apply pipeline
                case physics::OBS_EARTH:
                    synthAncil.time = anc.time;
                    synthAncil.detectorid = anc.detectorId;
                    synthAncil.scandir = anc.scanDir;
                    
                    // FUTURE: blackbody files should include temperatures over time
                    // this is a reasonable approximation for simulation since BBs are controlled to tens of millikelvin
                    calAncil.abbTemp = anc.abbTemp;
                    calAncil.hbbTemp = anc.hbbTemp;
                    calAncil.zbbTemp = anc.zbbTemp;
                    calAncil.reflectedTemp = anc.reflectedTemp; 
					
                    synth();
							
                    // FUTURE: load bbEmissivity from a bound CDF file properly interpolated to the wavenumbers
                    // currently just fills with 1.0 emissivities and transmission
                    // FUTURE: set from emiss port and pass in data through port? As of current sim data, simplified emissivity 
					// is appropriate
                    std::fill(calRef.bbEmissivity.begin(), calRef.bbEmissivity.end(), 1.0);
                    calRef.telescopeTau = ref.telescopeTau;
                    calRef.calmirrorTau = ref.calMirrorTau;                   
					// FUTURE: move this to	a reference data file (probably CDF) rather than generating it hard-coded
                    // FUTURE - replace with calculation (calRef's wavenumbers should be the effective, not target
                    // wnum scale).
                    calRef.wavenumbers = targetWnumScale; 
                    // MSO - set these in constructor for now

                    rpleAncil.laserWavenumberRatio = physics::lwnRatioFromTheta(ref.theta);
                    
                    c3b();
                    
                    rolloff();
                    resample();
                    break;
                default:
                    abort();             
                }
            }
        }; // class SimCal3
    
} //namespace gips

#endif
