#include "SimCal3Common.h"
#include <cassert>
#include <iostream> 

namespace gips{

    std::string outputFileNameFor( 
        const boost::filesystem::path &orig,   
        const boost::filesystem::path &target_dir,
        const char *modifier,
        const char *suffix)
    {
        using boost::filesystem::path;
        path dirname( orig.branch_path() );
        std::string filename( orig.leaf() );
        unsigned dotPos = filename.rfind('.');
        if (dotPos!=filename.npos)
            filename = filename.substr( 0, dotPos );
        return (target_dir / (filename + modifier + suffix)).string();
    };
    
    CdfFileOut::CdfFileOut( const std::string &filename_, unsigned lwSpcSize,
                            unsigned mwSpcSize, unsigned numRows, unsigned numCols): 
        filename(filename_)
    {
        ncf = std::auto_ptr<NcFile>( new NcFile( filename.c_str(), NcFile::Replace, 0, 0, NcFile::Offset64Bits ) ); 
        assert(ncf->is_valid());
        
        NcDim *npts_lw = ncf->add_dim("npts_lw", lwSpcSize);
        assert(npts_lw->is_valid());
        NcDim *npts_mw = ncf->add_dim("npts_mw", mwSpcSize);
        assert(npts_mw->is_valid());
        NcDim *row_indx = ncf->add_dim("row_indx", numRows);
        assert(row_indx->is_valid());
        NcDim *col_indx = ncf->add_dim("col_indx", numCols);
        assert(col_indx->is_valid());
        NcDim *time = ncf->add_dim("time", 1);
        assert(time->is_valid());
        
        // netcdf objects to hold return values;
        NcBool ncb;      
        NcVar *ncv;

        // create vars
        NcVar* obsTime = ncf->add_var("Obs_time", ncDouble, time);
        assert(obsTime->is_valid());
        ncb = obsTime->add_att("unit", "UTC hour");
        assert(ncb);
        ncv = ncf->add_var("signalReal_lw", ncFloat, 
                                  row_indx, col_indx, npts_lw);
        assert(ncv->is_valid());       
        ncv = ncf->add_var("signalReal_mw", ncFloat, 
                            row_indx, col_indx, npts_mw);
        assert(ncv->is_valid());       
        ncv = ncf->add_var("signalImag_lw", ncFloat, 
                            row_indx, col_indx, npts_lw);
        assert(ncv->is_valid());       
        ncv = ncf->add_var("signalImag_mw", ncFloat, 
                            row_indx, col_indx, npts_mw);
        assert(ncv->is_valid());       
        ncv = ncf->add_var("baseWnum", ncFloat, row_indx, col_indx);
        assert(ncv->is_valid());       
        ncv = ncf->add_var("deltaWnum", ncFloat, row_indx, col_indx);
        assert(ncv->is_valid());       
        
        // add generated_by global attribute
        const std::string tagName("$Id: SimCal3Common.cc,v 1.1.2.6 2006/03/22 22:18:56 rayg Exp $");
        ncb = ncf->add_att("generated_by", tagName.c_str());    
        assert(ncb);

        ncf->sync();  // flush
    };    

    CdfFileIn::CdfFileIn( const std::string &filename_, 
               physics::observation_type_e scenetype_):
        ncf( new NcFile( filename_.c_str(), NcFile::ReadOnly ) ),
        scenetype(scenetype_),
        filename(filename_)
    {
        // fetch time from the file, assuming one record per file
        // FUTURE: don't assume one record per file!
        
        NcError ncerr( NcError::verbose_nonfatal );
        
        epochTime = getDouble(std::string("Epoch_time"), 0.0);
        std::cout << "time for " << filename << ": " << epochTime << std::endl;
        
        obsTime = getDouble(std::string("Obs_time"), 0.0);
        telescopeTemp = getDouble(std::string("TeleTemp"), 0.0);
    }

    double CdfFileIn::getDouble(const std::string &varname, double defaultVal)
    {
        double d;
        NcVar *ncv = ncf->get_var( varname.c_str() );
        if (ncv!=NULL)
        {
            long count[] = { 1 };
            ncv->get( &d, &count[0] );
        }
        else
        {
            std::cerr << "WARNING: no " << varname << " in " << filename << ", setting to " << defaultVal << std::endl;
            d = defaultVal;
        }          
        return d;
    };
    
    CdfFileInSeq parseArgs( int argc, char *argv[] )
    {
        // options list must be pairs of options and values
        if(argc == 1 || argc%2 != 1) return CdfFileInSeq();
/*        {
            printUsageMsg(argv[0]);
            abort();
            }
*/       
        CdfFileInSeq files;
        int nH=0,nA=0,nZ=0,nE=0;
        
        for(int i=1; i<argc; i+=2)
        {
            switch( argv[i][0] )
            {
            case 'h':
            case 'H': 
                ++nH;
                files.push_back( CdfFileInSeq::value_type( new CdfFileIn( argv[i+1], physics::OBS_REF_UPPER ) ) ); 
                break;
                
            case 'w':
            case 'W': 
                ++nA;
                files.push_back( CdfFileInSeq::value_type( new CdfFileIn( argv[i+1], physics::OBS_REF_LOWER ) ) ); 
                break;
                
            case 's':
            case 'S': 
                ++nZ;
                files.push_back( CdfFileInSeq::value_type( new CdfFileIn( argv[i+1], physics::OBS_REF_SPACE ) ) );
                break;
                
                // FUTURE: o?            
            case 'e':
            case 'E':
                ++nE; 
                files.push_back( CdfFileInSeq::value_type( new CdfFileIn( argv[i+1], physics::OBS_EARTH ) ) ); 
                break;
                
            default:
//            printUsageMsg(argv[0]);
                std::cerr << "ERROR: unknown scene type '" << argv[i] << "', aborting\n";
//            abort();
                return CdfFileInSeq();
            }            
        }
        
        // encourage that we have matching number of blackbody views
        if ( (nH==0) || (nH!=nA) || (nH!=nZ) || (nE==0) )
        {
//            printUsageMsg(argv[0]);
            std::cerr << "ERROR: data should have equal number of blackbody views and at least one of each view; aborting\n";
            //           abort();
            return CdfFileInSeq();
        }
        return files;
    };   // parseArgs()
    
}// namespace gips
