/** 
*  \file SimCal3MpiTest.cc
*  Prototype GIPS pipeline designed to work on GIFTS Dec2005 simulation data.
*  This file contains code for a "quick and dirty" parallel executable wrapper around the SimCal3 pipeline.
*    
*  \author R.K.Garcia <rayg@ssec.wisc.edu>
*
*  \version $Id: SimCal3MpiTest.cc,v 1.1.4.21 2006/04/04 19:00:45 graemem Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* FUTURE: consider parallelizing file I/O. Possibly group readers and 
* writers separately within a Binder Group and specify no sharing of 
* resources among different binders to prevent race conditions. Binders
* can themselves be multi-threaded if desired.
*
*  \endverbatim
*/

// FLAGS
// #define DEBUG_TASK_DUMP /* write raw task data to binary file for use with readoutMain() */

#include <cassert>
#include <list>
#include <ctime> // time_t, time()
#include <algorithm>
#include <memory> // auto_ptr
#include <string>
#include <vector>
#include <iostream> 

// third-party boost libraries
#include <boost/array.hpp> // boost::array
//#include <boost/filesystem/path.hpp> // boost::filesystem::path

// MPI multiprocessing library
#include "mpi.h" // MPI_*()

#include "Physics.h"  // provides WavenumberScale<>
//#include "BindCdf.hxx" // provides a memory<->NetCDF API
#include "CplxUtils.hxx"  // oobVal<>()

#include "SimCal3Common.h" // consts, typedefs and routines common between SimCal3Test and SimCal3MpiTest

// uncomment for MPI_DEBUG_MODE: allow user to examine memory dump files
// by specifying mpi_size = 0
//define MPI_DEBUG_MODE

using namespace std;
using namespace gips;



//typedef boost::shared_ptr<FileToMemBinder> FileToMemBinder_P;
typedef int MPI_Result;

/////////////////////////////////////////////////////////
// local structs

//
// FUTURE: we're not using full MPI data types here, in favor of POD-blitting.
// In the case of a heterogeneous (little-endian / big-endian / different word size) cluster, we are SOL.
//

// FUTURE: physics::WnumScale<> is a pseudo-POD instead of a true POD; it has a constructor,
// but its data members are memcpy-friendly. If we want to be pedantic, however, incorrect!
struct slaveSettings_t // contains any pipeline settings for the slaves which need to be determined by file content
{
    LwSimCal3::settings_t lw;
    MwSimCal3::settings_t mw;
    
    // FUTURE: remove constructor from POD
    slaveSettings_t( const LwSimCal3::settings_t &lw_, const MwSimCal3::settings_t &mw_ ):
        lw(lw_), mw(mw_) { } 
};


// function to clear the content of an array to an out-of-bounds value
template< class ArrayType >
inline void array_clear( ArrayType &a )
{ 
    std::fill( a.begin(), a.end(), oobVal<typename ArrayType::value_type>() ); 
}

// 
// task blocks (1 or more tasks) are scattered to the slaves
//

struct task_input_t // is what goes on the wire to the slaves
{
    struct lw_input
    {
        LwSimCal3::control_t ctrl;
        LwIfgType ifg;
        LwSimCal3::ancillary_t ancil;
        LwSimCal3::ref_t ref;
        
        
        void clear()
        {
            ctrl = LwSimCal3::control_t();
            array_clear(ifg);
            ancil = LwSimCal3::ancillary_t(); 
            ref = LwSimCal3::ref_t();
        }
    } lw;

    struct mw_input
    {
        MwSimCal3::control_t ctrl;
        MwIfgType ifg;
        MwSimCal3::ancillary_t ancil;
        MwSimCal3::ref_t ref;

        void clear()
        {
            ctrl = MwSimCal3::control_t();
            array_clear(ifg);
            ancil = MwSimCal3::ancillary_t(); 
            ref = MwSimCal3::ref_t();
        }
    } mw;

    void clear()
    {
        lw.clear();
        mw.clear();
    }
};

struct task_output_t // is what the slaves return
{
    struct lw_output
    {
        LwRealSpectrumType spectrum, cal;
        LwCplxSpectrumType obsOut;
                    
        void clear()
        {
            array_clear(spectrum);
            array_clear(cal);
            array_clear(obsOut);
        }
    } lw;
    
    struct mw_output
    {
        MwRealSpectrumType spectrum, cal;
        MwCplxSpectrumType obsOut;
      
        void clear()
        {
            array_clear(spectrum);
            array_clear(cal);
            array_clear(obsOut);
       }
    } mw;
    // FUTURE: status output, monitoring output
    
    void clear()
    {
        lw.clear();
        mw.clear();
    }
};




/////////////////////////////////////////////////////////
// globals

int mpisize; // number of nodes in the computation
int mpirank; // rank of this node (0==master)


// one job == one data cube
// 16384 tasks per job
// blocks of tasks are sent to slaves

const unsigned tasks_per_block = 1; // number of tasks that'll be handed off at a time
const unsigned tasks_per_job = NUM_RECS;

// DEBUG: binary dump of task_input_ts
#ifdef DEBUG_TASK_DUMP
FILE *fpdebugin = NULL;
FILE *fpdebugout = NULL;
const char fndebugin[] = "test/task_input_t";
const char fndebugout[] = "test/task_output_t";
#endif



/////////////////////////////////////////////////////////
// master routines


slaveSettings_t loadSlaveSettings()
{
    // target wavenumber scales
    physics::WnumScale<double,LW_SPC_SIZE> lw_scale( physics::wnums_t::fromLaserWnum(LW_SPC_SIZE, LW_LWN, LW_DECIMATION ));
    physics::WnumScale<double,MW_SPC_SIZE> mw_scale( physics::wnums_t::fromLaserWnum(MW_SPC_SIZE, MW_LWN, MW_DECIMATION ));
    
    return slaveSettings_t
    ( 
        LwSimCal3::settings_t(lw_scale, LW_ROLLOFF_WNS[0], LW_ROLLOFF_WNS[1], 
                              LW_ROLLOFF_WNS[2], LW_ROLLOFF_WNS[3]),
        MwSimCal3::settings_t(mw_scale, MW_ROLLOFF_WNS[0], MW_ROLLOFF_WNS[1], 
                              MW_ROLLOFF_WNS[2], MW_ROLLOFF_WNS[3])
    );        
}

#define DEBUGVAL(x) x[x.size()/4]


template< typename BandInputType >
void completeMissingFields( BandInputType &x, const CdfFileIn &file, unsigned tasknum, unsigned reductionFactor )
{    
    x.ancil.obsType = file.scenetype;
    x.ancil.time = file.epochTime; 
    x.ancil.hbbTemp = HBB_TEMP;
    x.ancil.abbTemp = ABB_TEMP;
    x.ancil.zbbTemp = ZBB_TEMP;
    x.ancil.reflectedTemp = REFL_TEMP;
    x.ancil.detectorId = tasknum;
    x.ancil.scanDir = physics::SCAN_FWD;
    
    x.ref.theta = thetaFromDetectorId(tasknum); // FUTURE: retire
    x.ref.telescopeTau = TELESCOPE_TAU;
    x.ref.calMirrorTau = CAL_MIRROR_TAU;
}

// print command line usage info
void printUsageMsg(const char* argv0)
{
    std::cout 
        << "Usage: \n" 
        << "   mpirun -np X " << argv0 << " [filetype filename] [filetype filename] ...\n" 
        << "where\n"
        << "   X is the number of nodes, s.t. X=(2^n + 1) for some integer n>0\n"
        << "   filetype is one of the values:\n"
        << "      H     - hot blackbody data cube\n"
        << "      W     - warm blackbody data cube\n"
        << "      S     - space blackbody data cube\n"
        << "      E     - Earth observation data cube\n"
        << "Options list must specify at least one of each blackbody and at least one observation cube.\n";
}; 

void masterMain( int argc, char **argv )
{
#ifdef DEBUG_TASK_DUMP
    fpdebugin = fopen(fndebugin,"wb");
    fpdebugout = fopen(fndebugout,"wb");
#endif

    // grab information on input files
    CdfFileInSeq files( parseArgs( argc, argv ) );
    if (!files.size())
    {
        printUsageMsg(argv[0]);
        exit(5);
    }

    // FUTURE: replace this with more of a task iterator?
    const unsigned n_jobs = files.size();
    const unsigned begin = 0, end = tasks_per_job;
    const unsigned nslaves = mpisize-1;
    task_input_t todo;
    std::vector<MPI_Request> todo_request(nslaves);
    std::vector<task_output_t> done(nslaves);
    std::vector<MPI_Request> done_request(nslaves);
    std::vector<MPI_Status> done_status(nslaves);
    int rc;

    // broadcast iterations data (how many tasks to expect, total)
    // FUTURE: we assume that number of tasks evenly divides into number of slaves
    assert( (end-begin) % nslaves == 0 );
    unsigned tasks_per_slave = (end - begin)/nslaves * n_jobs;
    // AAAAAAAA
    MPI_Bcast( &tasks_per_slave, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD );

    // broadcast settings data (used to create pipelines)
    // for each input file
    // BBBBBBBB
    slaveSettings_t settings( loadSlaveSettings() ); 
    MPI_Bcast( &settings, sizeof(settings), MPI_BYTE, 0, MPI_COMM_WORLD );
    
    CdfFileOut *pFileOutput_unresampled = NULL; // we don't output for all files
    CdfFileOut *pFileOutput = NULL; 
    task_output_t todisk;

    for( CdfFileInSeq::const_iterator curf = files.begin(); curf!=files.end(); ++curf )
    {
        // bind the file contents to the task input buffer
        CdfFileIn &file( **curf );
        cout << "processing " << file.filename << endl;

        file.bindVar(todo.lw.ifg, string("signalReal_lw"));
        file.bindVar(todo.mw.ifg, string("signalReal_mw"));
        //fileIn.bindVar(lwReference.theta, string("Theta"));
        //fileIn.bindVar(mwReference.theta, string("Theta"));

        // if this is an earth cube, set up output pack to write output to file
        if (file.scenetype == physics::OBS_EARTH)
        {
            // clunky but effective
            string regular = outputFileNameFor( file.filename, "test", "_mpi" ); 
            string unresampled = outputFileNameFor( file.filename, "test", "_mpi_unresampled" );

            cout << "creating output files " << regular << " and " << unresampled << endl;

            pFileOutput_unresampled = new CdfFileOut( unresampled, 
                                        LW_SPC_SIZE, MW_SPC_SIZE, 
                                        NUM_ROWS, NUM_COLS);
            pFileOutput = new CdfFileOut( regular,
                                        LW_SPC_SIZE, MW_SPC_SIZE, 
                                        NUM_ROWS, NUM_COLS);

            pFileOutput->bindVar(todisk.lw.spectrum, string("signalReal_lw"));
            pFileOutput->bindVar(todisk.mw.spectrum, string("signalReal_mw"));
            pFileOutput_unresampled->bindVar(todisk.lw.cal, string("signalReal_lw"));
            pFileOutput_unresampled->bindVar(todisk.mw.cal, string("signalReal_mw"));

            // FUTURE: output baseWnum, deltaWnum
//        mem2disk.push_back( newCdfFieldWriter( &myBench.baseWnum, NcVar_p(fpOut->get_var(realName.c_str())), shape, shape+3, &mapping ));
        }

        for( unsigned tasknum=begin; tasknum!=end; )
        {   
            static time_t T = time(NULL) + 5;
            if (time(NULL)>T) 
            {
                T = time(NULL) + 5;
                cout << "working on task group starting with " << tasknum << endl;
            }

            for( unsigned slavenum=0; slavenum!=nslaves; ++slavenum, ++tasknum )
            { // this is effectively a scatter-gather with a single outgoing buffer
                file.nextRec();
                
                // fill in any fields not available in the file itself
                completeMissingFields( todo.lw, file, tasknum, LW_DECIMATION ); 
                completeMissingFields( todo.mw, file, tasknum, MW_DECIMATION ); 
                todo.lw.ctrl.operation = LwSimCal3::PROCESS_DATA;
                todo.mw.ctrl.operation = MwSimCal3::PROCESS_DATA;
                
                // cout << "master: task " << todo.lw.ancil.detectorId << ' ' << DEBUGVAL(todo.mw.ifg) << endl; // DEBUG
                
#ifdef DEBUG_TASK_DUMP
                if (fpdebugin) fwrite( &todo, sizeof(todo), 1, fpdebugin ); // DEBUG
#endif

                // CCCCCCCC
                rc = MPI_Issend( &todo, sizeof(todo), MPI_BYTE, 
                                     slavenum+1, todo.lw.ancil.detectorId, MPI_COMM_WORLD, &todo_request[slavenum] );
                assert(MPI_SUCCESS==rc);
                // FUTURE: consider whether to do this in the case of no output being needed?
                // DDDDDDDD
                rc = MPI_Irecv( &done[slavenum], sizeof(done[0]), MPI_BYTE, 
                                slavenum+1, MPI_ANY_TAG, MPI_COMM_WORLD, &done_request[slavenum] );
                assert(MPI_SUCCESS==rc);
            } // for each slave
            
            // wait for everything to complete (i.e. all 'done's to be local) before continuing
            // EEEEEEEE
            rc = MPI_Waitall(nslaves,&done_request[0],&done_status[0]);
            assert(MPI_SUCCESS == rc);
            
            if (pFileOutput_unresampled) 
            { // then write out done blocks to the output file
                for( unsigned i=0; i<nslaves; i++ )
                {
                    pFileOutput->nextRec();
                    pFileOutput_unresampled->nextRec();

                    todisk = done[i]; // copy to outgoing disk buffer; FUTURE: do without copy
                    //cout << "master: received " << DEBUGVAL(done[i].mw.spectrum) << endl; // DEBUGpFileOutput_unresampled->nextRec();
    
#ifdef DEBUG_TASK_DUMP
                    if (fpdebugout) fwrite( &todisk, sizeof(todisk), 1, fpdebugout );
#endif
                }
            }
            
        } // for each block of tasks
        
        if (pFileOutput_unresampled != NULL)
        {
            delete pFileOutput_unresampled; pFileOutput_unresampled = NULL;
            delete pFileOutput; pFileOutput = NULL;
        }
            
    } // for each file
#ifdef DEBUG_TASK_DUMP
    if (fpdebugin) fclose(fpdebugin);
    if (fpdebugout) fclose(fpdebugout);
#endif
}

void slaveMain() // runs on nodes with mpirank>0
{
    // receive iterations data
    // AAAAAAAA
    unsigned ntasks = 0;
    MPI_Bcast( &ntasks, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD );
    
    // receive settings data
    // BBBBBBBB
    // FUTURE: this is a non-obvious copy/constructor workaround which is only slightly better 
    // than using hardcoded values like the master uses.
    // FUTURE: consider solving this with a mixin template CopiedFromMpiBCast<PodType> 
    //         which offers explicit ctor from an MPI communicator and a rank?
    //  e.g. CopiedFromMpiBCast<settings_t> settings( 0, MPI_COMM_WORLD );

    slaveSettings_t *psettings = (slaveSettings_t *)malloc(sizeof(slaveSettings_t)); 
    slaveSettings_t &settings(*psettings);
    MPI_Bcast( &settings, sizeof(settings), MPI_BYTE, 0, MPI_COMM_WORLD );
    
    // task buffers, as simple as I dare make them
    task_input_t input;
    task_output_t output;

    // FUTURE: for real data, need real emissivities; 
    //         for current simulation data a simplified emissivity (or none at all) is appropriate
    static EmissType blankLwEmissivity, blankMwEmissivity;
    std::fill( blankLwEmissivity.begin(), blankLwEmissivity.end(), 1.0 );
    std::fill( blankMwEmissivity.begin(), blankMwEmissivity.end(), 1.0 );
    
    // create pipeline instances
    static LwSimCal3 lwpipe( settings.lw, 
                             LwSimCal3::Ports(&input.lw.ancil, &input.lw.ref, &input.lw.ctrl, 
                                              &input.lw.ifg, &blankLwEmissivity, &output.lw.obsOut,
                                              &output.lw.cal, &output.lw.spectrum));
//                        portsForPipeline<LwSimCal3::Ports>( &input.lw, &output.lw, &blankLwEmissivity ) );
    static MwSimCal3 mwpipe( settings.mw, 
                             MwSimCal3::Ports(&input.mw.ancil, &input.mw.ref, &input.mw.ctrl, 
                                              &input.mw.ifg, &blankMwEmissivity, &output.mw.obsOut, 
                                              &output.mw.cal, &output.mw.spectrum));
//                        portsForPipeline<MwSimCal3::Ports>( &input.mw, &output.mw, &blankMwEmissivity ) );

    int rc;
    
    for( unsigned t = 0; t<ntasks; ++t )
    {
        // FUTURE: we're wasting valuable CPU time with synchronous receive
        
        /// receive block of data from input service (currently == master)
        // CCCCCCCC
        MPI_Result res=0;
        MPI_Status status;
        res = MPI_Recv(&input, sizeof(input), MPI_BYTE, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);


        //// process task
        //compute_derived(input);
        output.clear();
        lwpipe();
        mwpipe();

        //cout << "rank " << mpirank << " processed task " << input.lw.ancil.detectorId 
        //        << ": " << DEBUGVAL(input.mw.ifg) << "->" << DEBUGVAL(output.mw.spectrum) << endl; // DEBUG
        // return results to the master
        // DDDDDDDD
        rc = MPI_Send(&output, sizeof(output), MPI_BYTE, 0, input.lw.ancil.detectorId, MPI_COMM_WORLD);
    }

    free(psettings); // FUTURE: get rid of this
}

#ifdef MPI_DEBUG_MODE
// Debug main runs when mpisize==1, i.e. run without MPI. 
// It reads through the test data dump so that we can page through the data using GDB, quick and easy.
void readoutMain( int argc, char *argv[] )
{
    fpdebugin = fopen( fndebugin, "rb" );
    fpdebugout = fopen( fndebugout, "rb" );
    task_input_t inp;
    task_output_t oup;
  
    for(unsigned n=0;!feof(fpdebugin) && !feof(fpdebugout);n++)
    {
        fread( &inp, sizeof(inp), 1, fpdebugin);
        fread( &oup, sizeof(oup), 1, fpdebugout);
        std::cout << n << '\r'; std::cout.flush();
    }
}
#endif

int main( int argc, char *argv[] )
{

//    slaveSettings_t settings = loadSlaveSettings();

    MPI_Init( &argc, &argv );

    MPI_Comm_size(MPI_COMM_WORLD,&mpisize);
    MPI_Comm_rank(MPI_COMM_WORLD,&mpirank);

#ifdef MPI_DEBUG_MODE
    if (mpisize==1) 
        {
            readoutMain(argc,argv);
            return 0;
        }
#endif
    
    switch( mpirank )
    {
        case 0: masterMain( argc, argv ); break;
        default: slaveMain(); break;
    }
    return 0;  
}



