/** 
  *  \file CosineRolloff.hxx
  *  \brief Cosine data rolloff capsule
  *  \verbatim
  * OVERVIEW
  * 
  * This capsule applies a sinusoidal rolloff to the input data array using 
  * indices p0, p1, p2 and p3, as follows:
  * 
  *   out[0:p0]   values set to 0
  *   out[p0:p1]  cosine rolloff, calculated s.t. out[p0]=0 and out[p1]=in[p1]
  *   out[p1:p2]  set to in[p1:p2]
  *   out[p2:p3]  cosine rolloff, calculated s.t. out[p2]=in[p2] and out[p3]=0
  *   out[p3:end] values set to 0
  *
  * A slight performance gain can be realized by rolling off in place, ie. 
  * passing in the same buffer for input and output. If the addresses of the 
  * and output data arrays are the same, the copies of values p1 through p2 are 
  * skipped. 
  *
  * INTERFACE
  *
  *   Ports:
  *   -- anc_in: ancillary data struct (per-invocation data)
  *   -- in: input data buffer
  *   -- out: output data buffer
  *
  *   Ancillary:
  *   -- p0, p1, p2, p3: indices of rolloff points, used as described above 
  *
  * TEMPLATE SPECIFICATION
  *
  *   Parameters:
  *   -- InType: type of buffer "in" 
  *   -- OutType: type of buffer "out"
  *
  *   Requirements:
  *   -- InType and OutType must be boost-array-like types with members (as 
  *      described in boost::array documentation): begin(), end(), size(), 
  *      value_type, static_size
  *   -- InType::value_type and OutType::value_type must be inter-compatible 
  *      and compatible with type int, and must support arithmetic operators 
  *      ('/', '*', '+', '-') 
  *   -- InType::static_size == OutType::static_size
  *  \endverbatim
  *
  *  \author G. Martin <graemem@ssec.wisc.edu>
  *
  *  \version $Id: CosineRolloff.hxx,v 1.2.2.15 2006/04/04 19:00:45 graemem Exp $
  *  \par Copyright:
  *  \verbatim
  *
  *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2006
  *  Space Science and Engineering Center
  *  University of Wisconsin - Madison, USA
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
  * \endverbatim
  *
  *
  */
#ifndef HXX_COSINEROLLOFF
#define HXX_COSINEROLLOFF

#include <iterator>
#include <boost/static_assert.hpp>

namespace gips 
{
    template <typename InType, typename OutType>
    class CosineRolloff
    {
    public:

        BOOST_STATIC_ASSERT((unsigned(InType::static_size) == unsigned(OutType::static_size)));

        struct ancillary_t
        {
            unsigned p0;
            unsigned p1;
            unsigned p2; 
            unsigned p3;
        };
        struct Ports
        {
            const ancillary_t *anc_in;
            const InType *in;
            OutType *out;
            Ports( const ancillary_t *anc_in_, const InType *in_, OutType *out_): 
                  anc_in(anc_in_), in(in_), out(out_) {}
        };

    private:
        
        const Ports ports;

        // Precondition: iterator + range must be within bounds.
        // Apply sinusoidal rolloff to zero, over 'range' values. *it will not
        // be changed, and *(it + range) will be 0. Sign of range indicates rolloff 
        // direction.
        void rolloffSinusoidal(typename OutType::iterator it, int range)
            {       
                typename OutType::value_type amplitudeRatio = *it/ 2.0; 

                if(range<0)
                {
                    --it;
                    for(int i=-1; i>=range; --i, --it)
                        *it = (std::cos((i*M_PI)/range) + 1) * amplitudeRatio;
                }
                else
                {
                    ++it;
                    for(int i=1; i<=range; ++i, ++it)
                        *it = (std::cos((i*M_PI)/range) + 1) * amplitudeRatio;
                }
            }
      
        // return true if address of x == address of y
        template<typename A, typename B>
        static bool sourceIsTarget( const A &x, const B& y ) 
            {
                return false; // different in and out yypes cannot have same address
            }
        template<typename A>
        static bool sourceIsTarget( const A &x, const A &y )
            { 
                return bool( &x == &y );
            }
           

    public:
        
        CosineRolloff(const Ports &p_): ports(p_) {}   // ctor
        ~CosineRolloff() {}                            // dtor
        void operator()(void)                          // ftor
            {
                // convenience refs
                const unsigned &p0(ports.anc_in->p0);
                const unsigned &p1(ports.anc_in->p1);
                const unsigned &p2(ports.anc_in->p2);
                const unsigned &p3(ports.anc_in->p3);
                const InType &obsIn (*(ports.in)); 
                OutType &obsOut (*(ports.out));
             
                // FUTURE: replace with exception
                // check indices are sequential
                assert((p0 < p1) && (p1 < p2) && (p2 < p3));
                
                // FUTURE: replace with exception
                // check indices within bounds
                assert(p3 < obsIn.size());
                
                // zero values before p0
                std::fill(obsOut.begin(), obsOut.begin() + p0, typename OutType::value_type(0));

                // copy values [p1:p2] if input and output buffers are different
                if (!sourceIsTarget(obsIn,obsOut))
                    std::copy(obsIn.begin() + p1, obsIn.begin() + p2 + 1, obsOut.begin() + p1); 

                // generate sin from p0 to p1
                rolloffSinusoidal(obsOut.begin() + p1, int(p0) - int(p1));
               
                // generate sin from p2 to p3
                rolloffSinusoidal(obsOut.begin() + p2, int(p3) - int(p2));
                
                // zero values after p3
                std::fill(obsOut.begin() + p3 + 1, obsOut.end(), typename OutType::value_type(0));                
            }
    };
    
} // namespace gips

#endif
