/** 
  *  \file Resample.hxx
  *  \brief Spectral Resampling Algorithm.
  *
  *  \author M. Smuga-Otto <maciek@ssec.wisc.edu>
  *
  *  \version $Id: Resample.hxx,v 1.7.2.8 2006/03/30 21:23:26 graemem Exp $
  *
  *  \par Copyright:
  *  \verbatim
  *
  *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004-2006
  *  Space Science and Engineering Center
  *  University of Wisconsin - Madison, USA
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
  * \endverbatim
  */

#ifndef HXX_RESAMPLE
#define HXX_RESAMPLE

// C++ STL includes
#include <algorithm>
#include <cmath>
#include <functional>
#include <boost/array.hpp>



// GIPS capsule includes
#include "CplxFFT3.hxx"
#include "Physics.h"

#ifdef DEBUG_FBF_TMP_OUT
#include "FBF.h"
#include "CplxUtils.hxx"   
#endif

namespace gips {

        
template <typename InType, typename OutType, std::size_t UPSAMPLED_SIZE>
class SpectralResampling
{
public:
    typedef physics::WnumScale<double, InType::static_size> WavenumberScale;
    static WavenumberScale nullwn;
    struct settings_t
    {
        WavenumberScale wavenumberScale; 

        settings_t(WavenumberScale w_ = nullwn): 
            wavenumberScale(w_) {} // invokes physics::WnumScale copy constructor
    };
    struct ancillary_t
    {
        double laserWavenumberRatio;
    };
    struct Ports
    {
        const ancillary_t * anc_in;
        const InType * obs_in;
        OutType * obs_out;
        Ports(const ancillary_t *anc_in_, const InType *obs_in_, OutType *obs_out_): 
             anc_in(anc_in_), obs_in(obs_in_), obs_out(obs_out_) {}
    };
private:
    typedef boost::array<std::complex<double>, InType::static_size * 2> SmallBuffer; 
    typedef boost::array<std::complex<double>, UPSAMPLED_SIZE> BigBuffer; 
    typedef CplxFFT<SmallBuffer, SmallBuffer> SmallFFT; 
    typedef CplxFFT<BigBuffer, BigBuffer> BigFFT; 
    
    const settings_t settings;
    const Ports ports;
    const double normalization;
    SmallBuffer mirroredSpectrum, shortIfg;
    BigBuffer zeroFilledIfg, upsampledSpectrum;

    SmallFFT forwardFFT;
    BigFFT  backwardFFT; 

    template<typename FFT>
    static typename FFT::settings_t getFFTSettings(
                        typename FFT::direction_e d,
                        typename FFT::training_e t)
    {
        typename FFT::settings_t s;
        s.fftDirection = d;
        s.fftTraining = t;
        return s;
    }
        

public:
    SpectralResampling(const settings_t &s_, const Ports &p_): settings(s_), ports(p_),
         normalization(1.0 / static_cast<double>(ports.obs_in->size() * 2)),
         forwardFFT(
                 getFFTSettings<SmallFFT>( SmallFFT::FORWARD, SmallFFT::MEASURE),  
                 SmallFFT::Ports(&mirroredSpectrum, &shortIfg)),
         backwardFFT(
                 getFFTSettings<BigFFT>( BigFFT::BACKWARD, BigFFT::MEASURE),
                 BigFFT::Ports(&zeroFilledIfg, &upsampledSpectrum))
    {
        std::fill(zeroFilledIfg.begin(), zeroFilledIfg.end(), 0.0);
    } 

    ~SpectralResampling() {}
    void operator()() 
    {
        const InType &obsIn (*(ports.obs_in)); // avoid convoluted pointer deference syntax in subsequent client code.

        std::copy(obsIn.begin(), obsIn.end(), mirroredSpectrum.begin());
        std::copy(obsIn.begin(), obsIn.end(), mirroredSpectrum.rbegin()); 

        forwardFFT(); // reap benefits of CplxFFT capsule

        const typename SmallBuffer::iterator shortIfgMiddle(
            shortIfg.begin() + ports.obs_in->size() 
        );
        const typename SmallBuffer::reverse_iterator shortIfgRMiddle(
            shortIfg.rbegin() + ports.obs_in->size() 
        );
        

        std::copy(shortIfg.begin(), shortIfgMiddle, zeroFilledIfg.begin());
        std::copy(shortIfg.rbegin(), shortIfgRMiddle, zeroFilledIfg.rbegin());

        backwardFFT();

#ifdef DEBUG_FBF_TMP_OUT
        {
            typedef boost::array<double, 262144>  BigReal;
            BigReal* debugvec = new BigReal;
            extractComplexAbs(upsampledSpectrum, *debugvec);
            FBF::save("upsampledAbs.real8.262144",  *debugvec);
            extractComplexPhase(upsampledSpectrum, *debugvec);
            FBF::save("upsampledPhase.real8.262144",  *debugvec);
            delete debugvec;
        }
#endif
        
        unsigned upsampledSpectrumSize = UPSAMPLED_SIZE / 2; 

        const typename BigBuffer::iterator upsampledSpectrumMiddle(
            upsampledSpectrum.begin() + upsampledSpectrumSize
        );
        std::transform
        (
            upsampledSpectrum.begin(), 
            upsampledSpectrumMiddle, 
            upsampledSpectrum.begin(),
            std::bind2nd(std::multiplies<std::complex<double> >(), normalization)
        );

        
        // downsample:
        double deltaB = settings.wavenumberScale.asPOD().deltaWnum;

        double lwr(ports.anc_in->laserWavenumberRatio); // for brevity
//        std::cout << "lwr: " << lwr << std::endl; // DEBUG
        double rcoeff  = double(upsampledSpectrumSize) / (ports.obs_in->size() * deltaB * lwr);
        double roffset = settings.wavenumberScale[0] * (1 - lwr);
        
        OutType &obsOut (*(ports.obs_out));
        for (unsigned int k = 0; k < ports.obs_in->size(); ++k)
        {
            double rstep = deltaB * k;
            double rk = rcoeff * (roffset + rstep); // corresponding index in upsampled spectrum
                        
            if ((rk < 0) || (rk >= upsampledSpectrumSize - 1))
                obsOut[k] = 0;
            else
            {
                int rki = static_cast<int>(floor(rk)); // integral part
                double rkf = rk - rki;                   // fractional part
                obsOut[k] = std::real(upsampledSpectrum[rki])
                        + rkf * (std::real(upsampledSpectrum[rki+1]) 
                        - std::real(upsampledSpectrum[rki]));
            }
        }
    } // end operator()
}; // end class SpectralResampling
} // end namespace gips
#endif

